@@ -1211,7 +1211,6 @@ static void read_input(StringBuilder *buf)
   size_t lplen = (size_t)ml_get_len(lnum);
 
   while (true) {
-    lplen -= written;
     if (lplen == 0) {
       len = 0;
     } else if (lp[written] == NL) {
@@ -1220,11 +1219,11 @@ static void read_input(StringBuilder *buf)
       kv_push(*buf, NUL);
     } else {
       char *s = vim_strchr(lp + written, NL);
-      len = s == NULL ? lplen : (size_t)(s - (lp + written));
+      len = s == NULL ? lplen - written : (size_t)(s - (lp + written));
       kv_concat_len(*buf, lp + written, len);
     }
 
-    if (len == lplen) {
+    if (len == lplen - written) {
       // Finished a line, add a NL, unless this line should not have one.
       if (lnum != curbuf->b_op_end.lnum
           || (!curbuf->b_p_bin && curbuf->b_p_fixeol)
